/*
 *   GRacer
 *
 *   Copyright (C) 1999 Takashi Matsuda <matsu@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <unistd.h>
#include <string.h>
#include <math.h>
#include <common/gr_global.h>
#include <common/gr_memory.h>
#include "ac2scene.h"

typedef enum {
  GR_SCENE_OPT_TRIANGLE_STRIP_START,
  GR_SCENE_OPT_TRIANGLE_STRIP_LEFT,
  GR_SCENE_OPT_TRIANGLE_STRIP_RIGHT,

  GR_SCENE_OPT_TRIANGLE_FAN_START,
  GR_SCENE_OPT_TRIANGLE_FAN_LOOP,

  GR_SCENE_OPT_QUAD_STRIP_START,
  GR_SCENE_OPT_QUAD_STRIP_LOOP,

} GrSceneOptState;

typedef struct {
  GLenum type;
  GrSceneOptState state;
  int left, right, other;
  GrVertex left_v, right_v, other_v;
  GrSurface *surf;
} GrSceneOptHint;

static GrScene *	ac2gr_scene	(Ac3dScene *ascene, char *cwd);
static GrObject *	ac2gr_object	(GrScene *scene,
    					 Ac3dObject *aobj,
					 char *cwd);

static int
count_triangle_surfaces (GrObject *object)
{
  int i;
  int num = 0;
  GrSurface *surf;

  for (i=0; i<object->num_surfs; i++) {
    surf = object->surfs[i];
    if (surf->type == GL_POLYGON && surf->num_elements == 3) {
      num ++;
    }
  }

  return num;
}

static int
count_quad_surfaces (GrObject *object)
{
  int i;
  int num = 0;
  GrSurface *surf;

  for (i=0; i<object->num_surfs; i++) {
    surf = object->surfs[i];
    if (surf->type == GL_POLYGON && surf->num_elements == 4) {
      num ++;
    }
  }

  return num;
}


static int
check_triangle_strip (GrSceneOptHint *hint, GrSurface *surf)
{
  int i;
  int index;
  int left, right, other;

  if (surf->num_elements != 3)
    return 0;

  switch (hint->state) {
  case GR_SCENE_OPT_TRIANGLE_STRIP_START:
    for (i=0; i<3; i++) {
      if (surf->elements[i].ref == hint->other &&
	  surf->elements[(i+1)%3].ref == hint->left &&
	  surf->elements[(i+2)%3].ref == hint->right) {
	break;
      }
    }
    if (i != 3) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      other = i;
      left = (i+1) % 3;
      right = (i+2) % 3;

      hint->surf->num_elements = 3;

      hint->surf->elements[0] = surf->elements[left];
      hint->surf->elements[1] = surf->elements[right];
      hint->surf->elements[2] = surf->elements[other];

      hint->left = surf->elements[other].ref;
      hint->state = GR_SCENE_OPT_TRIANGLE_STRIP_RIGHT;

      return 1;
    }
    break;

  case GR_SCENE_OPT_TRIANGLE_STRIP_RIGHT:
  case GR_SCENE_OPT_TRIANGLE_STRIP_LEFT:
    for (i=0; i<3; i++) {
      if (surf->elements[i].ref == hint->left &&
	  surf->elements[(i+1)%3].ref == hint->right) {
	break;
      }
    }
    if (i != 3) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      left = i;
      right = (i+1) % 3;
      other = (i+2) % 3;

      index = hint->surf->num_elements++;
      hint->surf->elements[index] = surf->elements[other];

      if (hint->state == GR_SCENE_OPT_TRIANGLE_STRIP_RIGHT) {
	hint->right = surf->elements[other].ref;
	hint->state = GR_SCENE_OPT_TRIANGLE_STRIP_LEFT;
      } else {
	hint->left = surf->elements[other].ref;
	hint->state = GR_SCENE_OPT_TRIANGLE_STRIP_RIGHT;
      }
      return 1;
    }
    break;

  default:
  }

  return 0;
}

static int
check_triangle_fan (GrSceneOptHint *hint, GrSurface *surf)
{
  int i;
  int index;
  int left, right, other;

  if (surf->num_elements != 3)
    return 0;

  switch (hint->state) {
  case GR_SCENE_OPT_TRIANGLE_FAN_START:
    for (i=0; i<3; i++) {
      if (surf->elements[i].ref == hint->left &&
	  surf->elements[(i+1)%3].ref == hint->right &&
	  surf->elements[(i+2)%3].ref == hint->other) {
	break;
      }
    }
    if (i != 3) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      left = i;
      right = (i+1) % 3;
      other = (i+2) % 3;

      hint->surf->num_elements = 3;

      hint->surf->elements[0] = surf->elements[left];
      hint->surf->elements[1] = surf->elements[right];
      hint->surf->elements[2] = surf->elements[other];

      hint->right = surf->elements[other].ref;
      hint->state = GR_SCENE_OPT_TRIANGLE_FAN_LOOP;

      return 1;
    }
    break;

  case GR_SCENE_OPT_TRIANGLE_FAN_LOOP:
    for (i=0; i<3; i++) {
      if (surf->elements[i].ref == hint->left &&
	  surf->elements[(i+1)%3].ref == hint->right) {
	break;
      }
    }
    if (i != 3) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      other = (i+2) % 3;
      index = hint->surf->num_elements++;

      hint->surf->elements[index] = surf->elements[other];

      hint->right = surf->elements[other].ref;

      return 1;
    }
    break;

  default:
  }

  return 0;
}

static int
check_quad_strip (GrSceneOptHint *hint, GrSurface *surf)
{
  int i;
  int index;
  int left, right, other, remain;

  if (surf->num_elements != 4)
    return 0;

  switch (hint->state) {
  case GR_SCENE_OPT_QUAD_STRIP_START:
    for (i=0; i<4; i++) {
      if (surf->elements[i].ref == hint->other &&
	  surf->elements[(i+1)%4].ref == hint->left &&
	  surf->elements[(i+2)%4].ref == hint->right) {
	break;
      }
    }
    if (i != 4) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      other = i;
      left = (i+1) % 4;
      right = (i+2) % 4;
      remain = (i+3) % 4;

      hint->surf->num_elements = 4;

      hint->surf->elements[0] = surf->elements[left];
      hint->surf->elements[1] = surf->elements[right];
      hint->surf->elements[2] = surf->elements[other];
      hint->surf->elements[3] = surf->elements[remain];

      hint->left = surf->elements[other].ref;
      hint->right = surf->elements[remain].ref;
      hint->state = GR_SCENE_OPT_QUAD_STRIP_LOOP;

      return 1;
    }
    break;

  case GR_SCENE_OPT_QUAD_STRIP_LOOP:
    for (i=0; i<4; i++) {
      if (surf->elements[i].ref == hint->left &&
	  surf->elements[(i+1)%4].ref == hint->right) {
	break;
      }
    }
    if (i != 4) {
      /* found! */

      hint->surf->twoside = surf->twoside;
      hint->surf->smooth = surf->smooth;

      left = i;
      right = (i+1) % 4;
      remain = (i+2) % 4;
      other = (i+3) % 4;

      index = hint->surf->num_elements;

      hint->surf->num_elements += 2;

      hint->surf->elements[index] = surf->elements[other];
      hint->surf->elements[index+1] = surf->elements[remain];

      hint->left = surf->elements[other].ref;
      hint->right = surf->elements[remain].ref;

      return 1;
    }
    break;
  default:
  }

  return 0;
}



static int
find_vertex (GrObject *obj, GrVertex *v)
{
  int i;
  GrVertex w;

  for (i=0; i<obj->num_verts; i++) {
    w.c.x = X(obj->vertex, i);
    w.c.y = Y(obj->vertex, i);
    w.c.z = Z(obj->vertex, i);

    if (gr_vertex_near (v, &w, 0.001)) {
      return i;
    }
  }
  return -1;
}

static void
optimize_object (GrObject *obj,
    		 GrSceneOptHint **hints,
		 int num_hints)
{
  int find;
  int i, j;
  int num_triangles;
  int num_quads;
  int res;
  GrVertex left_v, right_v, other_v;
  GrSurface *surf;

  num_triangles = count_triangle_surfaces (obj);
  num_quads = count_quad_surfaces (obj);

  if (num_triangles == 0 && num_quads == 0) {
    return;
  }

  for (i=0; i<num_hints; i++) {
    gr_vertex_sub (&left_v, &hints[i]->left_v, &obj->pos);
    gr_vertex_sub (&right_v, &hints[i]->right_v, &obj->pos);
    gr_vertex_sub (&other_v, &hints[i]->other_v, &obj->pos);

    hints[i]->left = find_vertex (obj, &left_v);
    hints[i]->right = find_vertex (obj, &right_v);
    hints[i]->other = find_vertex (obj, &other_v);

    if (hints[i]->left == -1 ||
	hints[i]->right == -1 ||
	hints[i]->other == -1)
    {
      hints[i]->surf = NULL;
      continue;
    }

    hints[i]->surf = gr_new0 (GrSurface, 1);
    hints[i]->surf->type = hints[i]->type;

    switch (hints[i]->type) {
    case GL_TRIANGLE_STRIP:
      hints[i]->state = GR_SCENE_OPT_TRIANGLE_STRIP_START;
      hints[i]->surf->elements = gr_new (GrSElement, num_triangles + 2);
      break;

    case GL_TRIANGLE_FAN:
      hints[i]->state = GR_SCENE_OPT_TRIANGLE_FAN_START;
      hints[i]->surf->elements = gr_new (GrSElement, num_triangles + 2);
      break;


    case GL_QUAD_STRIP:
      hints[i]->state = GR_SCENE_OPT_QUAD_STRIP_START;
      hints[i]->surf->elements = gr_new (GrSElement, num_quads * 2 + 2);
      break;

    default:
    }
  }

  do {
    find = 0;

    for (i=0; i<num_hints; i++) {
      if (!hints[i]->surf)
	continue;

      for (j=0; j<obj->num_surfs; j++) {
	switch (hints[i]->type) {
	case GL_TRIANGLE_STRIP:
	  res = check_triangle_strip (hints[i], obj->surfs[j]);
	  break;

	case GL_TRIANGLE_FAN:
	  res = check_triangle_fan (hints[i], obj->surfs[j]);
	  break;

	case GL_QUAD_STRIP:
	  res = check_quad_strip (hints[i], obj->surfs[j]);
	  break;

	default:
	  res = 0;
	}
	if (res) {
	  find = 1;

	  gr_surface_free (obj->surfs[j]);
	  obj->surfs[j] = obj->surfs[--obj->num_surfs];
	}
      }
    }
  } while (find);

  for (i=0; i<num_hints; i++) {
    if (!(surf = hints[i]->surf))
      continue;

    if (surf->num_elements > 0) {
      surf->elements = gr_renew (surf->elements,
				 GrSElement,
				 surf->num_elements);

      obj->surfs[obj->num_surfs++] = surf;

    } else {
      gr_surface_free (hints[i]->surf);
    }
  }
}

void
gr_object_optimize (GrObject *obj, int recursive)
{
  int i;
  GrSList *hint_list = NULL;
  int num_hints = 0;
  GrSceneOptHint *hint, **hints;
  GrSurface *surf;
  GLenum type;
  GrObject *kid;

  for (i=0; i<obj->num_kids; i++) {
    kid = obj->kids[i];

    if (kid->name && kid->num_surfs == 1 && kid->surfs[0]->num_elements >= 3)
    {
      if (!strcmp (kid->name, "triangle_strip")) {
	type = GL_TRIANGLE_STRIP;
      } else if (!strcmp (kid->name, "triangle_fan")) {
	type = GL_TRIANGLE_FAN;
      } else if (!strcmp (kid->name, "quad_strip")) {
	type = GL_QUAD_STRIP;
      } else {
	type = GL_NONE;
      }

      if (type != GL_NONE) {
	hint = gr_new0 (GrSceneOptHint, 1);
	hint->type = type;

	surf = kid->surfs[0];

	hint->left_v.c.x = X(kid->vertex, surf->elements[0].ref);
	hint->left_v.c.y = Y(kid->vertex, surf->elements[0].ref);
	hint->left_v.c.z = Z(kid->vertex, surf->elements[0].ref);

	hint->right_v.c.x = X(kid->vertex, surf->elements[1].ref);
	hint->right_v.c.y = Y(kid->vertex, surf->elements[1].ref);
	hint->right_v.c.z = Z(kid->vertex, surf->elements[1].ref);

	hint->other_v.c.x = X(kid->vertex, surf->elements[2].ref);
	hint->other_v.c.y = Y(kid->vertex, surf->elements[2].ref);
	hint->other_v.c.z = Z(kid->vertex, surf->elements[2].ref);

	gr_vertex_add_a (&hint->left_v, &kid->pos);
	gr_vertex_add_a (&hint->right_v, &kid->pos);
	gr_vertex_add_a (&hint->other_v, &kid->pos);

	hint_list = gr_slist_prepend (hint_list, hint);

	gr_DECREF (kid);
	obj->kids[i--] = obj->kids[--obj->num_kids];

	num_hints ++;
      }
    }
  }

  if (num_hints > 0) {
    hints = gr_new (GrSceneOptHint *, num_hints);
    
    for (i=0; i<num_hints; i++) {
      hints[i] = hint_list->data;
      hint_list = gr_slist_remove (hint_list);
    }

    for (i=0; i<obj->num_kids; i++) {
      optimize_object (obj->kids[i], hints, num_hints);
    }

    for (i=0; i<num_hints; i++) {
      free (hints[i]);
    }
    free (hints);
  }

  if (recursive) {
    for (i=0; i<obj->num_kids; i++) {
      gr_object_optimize (obj->kids[i], 1);
    }
  }
}

void
gr_scene_optimize (GrScene *scene)
{
  int i;

  if (!scene)
    return;

  for (i=0; i<scene->num_objs; i++) {
    gr_object_optimize (scene->objs[i], 1);
  }
}

static void usage (FILE *file)
{
  fprintf (file,
       "Usage: following options are acceptable.\n"
       "	-i <filename>		specify ac3d scene file.\n"
       "	-o <filename>		output course file name.\n"
       "	-[xyz] <scale>		scale.\n"
       "	-t <threshold>		max distance between identical vertices.\n"
       "	-h			show this message.\n"
       );
}

int main (int argc, char *argv[])
{
  char *optstring = "i:o:x:y:z:t:h";
  char *iname = NULL;
  char *oname = NULL;
  FILE *ifile;
  FILE *ofile;
  Ac3dScene *ascene;
  GrScene *gscene;
  int c;
  float scale[3] = {10, 10, 10};
  float threshold = 0.01;
  int i;
  char *cwd = getcwd(NULL, 0);


  while ((c = getopt(argc, argv, optstring)) != EOF) {
    switch (c) {
      case 'i':
	iname = optarg;
	break;

      case 'o':
	oname = optarg;
	break;

      case 'x':
	sscanf (optarg, "%f", &scale[0]);
	break;

      case 'y':
	sscanf (optarg, "%f", &scale[1]);
	break;

      case 'z':
	sscanf (optarg, "%f", &scale[2]);
	break;

      case 't':
	sscanf (optarg, "%f", &threshold);
	break;

      case 'h':
	usage (stdout);
	return 0;

      default:
	usage (stderr);
	return -1;
    }
  }

  if (iname == NULL) {
    ifile = stdin;
  } else {
    ifile = fopen (iname, "r");
    if (!ifile) {
      fprintf (stderr, "couldn't open file \"%s\"", iname);
      return -1;
    }
  }

  if (oname == NULL) {
    ofile = stdout;
  } else {
    ofile = fopen (oname, "w");
    if (!ofile) {
      fprintf (stderr, "couldn't open file \"%s\"", oname);
      return -1;
    }
  }

  ascene = ac3d_scene_new_from_file (ifile);
  fclose (ifile);

  if (!ascene) {
    fprintf (stderr, "failed to read ac3d scene file.\n");
    return -1;
  }

  /* copy ac3d scene structures into gr_scene. */

  for (i=0; i<ascene->num_objs; i++) {
    ac3d_object_scale (ascene->objs[i], scale);
    ac3d_object_convert_coord (ascene->objs[i]);
  }
  gscene = ac2gr_scene (ascene, cwd);
  gr_scene_optimize (gscene);

  gr_scene_write_file (gscene, ofile);

  return 0;
}

static char *
relative_path (char *path, char *cwd)
{
  int len;

  if (!cwd)
    return path;

  len = strlen (cwd);
  if (len < 1) {
    return path;
  }

  if (!strncmp (cwd, path, len)) {
    if (cwd[len-1] == '/') {
      return path + len;
    } else {
      return path + len + 1;
    }
  } else {
    return path;
  }
}

static GrObject *
ac2gr_object (GrScene *scene, Ac3dObject *aobj, char *cwd)
{
  GrObject *obj;
  GrSurface *surf;
  int i, j;
  int type;

  obj = gr_new (GrObject, 1);
  if (!obj)
    return NULL;
  memset (obj, 0, sizeof (*obj));

  if (aobj->name && aobj->name[0] != '\0') {
    obj->name = aobj->name;
  }

  if (aobj->url && aobj->url[0] != '\0') {
    obj->url = aobj->url;
  }

  if (aobj->tex_name && aobj->tex_name[0] != '\0') {
    if (cwd) {
      obj->texture_name = relative_path (aobj->tex_name, cwd);
    } else {
      obj->texture_name = aobj->tex_name;
    }
  }

  if (aobj->options & AC3D_OPT_LOC) {
    obj->pos.c.x = aobj->loc[0];
    obj->pos.c.y = aobj->loc[1];
    obj->pos.c.z = aobj->loc[2];
  }

  obj->num_verts = aobj->num_verts;
  if (obj->num_verts > 0) {
    obj->vertex = gr_new (GLfloat, obj->num_verts * 3);
    for (i=0; i<obj->num_verts; i++) {
      X(obj->vertex, i) = X(aobj->verts, i);
      Y(obj->vertex, i) = Y(aobj->verts, i);
      Z(obj->vertex, i) = Z(aobj->verts, i);
    }
  }

  obj->num_surfs = aobj->num_surfs;
  if (obj->num_surfs > 0) {
    obj->surfs = gr_new (GrSurface*, obj->num_surfs);
    for (i=0; i<obj->num_surfs; i++) {
      surf = gr_new (GrSurface, 1);
      obj->surfs[i] = surf;
      memset (surf, 0, sizeof (GrSurface));
      type = aobj->surfs[i]->type & AC3D_SURF_TYPE_MASK;
      switch (type) {
      case AC3D_SURF_POLYGON:
	surf->type = GL_POLYGON;
	break;

      case AC3D_SURF_CLOSEDLINE:
	surf->type = GL_LINE_STRIP;
	break;

      case AC3D_SURF_LINE:
	surf->type = GL_LINE_STRIP;
	break;

      default:
	surf->type = GL_NONE;
      }
      surf->twoside = aobj->surfs[i]->type & AC3D_SURF_TWOSIDE;
      surf->smooth = aobj->surfs[i]->type & AC3D_SURF_SHADED;
      surf->num_elements = aobj->surfs[i]->num_refs;
      if (type == AC3D_SURF_CLOSEDLINE) {
	surf->num_elements ++;
      }
      surf->elements = gr_new (GrSElement, surf->num_elements);

      for (j=0; j<aobj->surfs[i]->num_refs; j++) {
	surf->elements[j].ref = aobj->surfs[i]->refs[j].ref;
	surf->elements[j].mat = aobj->surfs[i]->mat;
	surf->elements[j].s = aobj->surfs[i]->refs[j].s;
	surf->elements[j].t = aobj->surfs[i]->refs[j].t;
      }
      if (type == AC3D_SURF_CLOSEDLINE) {
	surf->elements[surf->num_elements-1] = surf->elements[0];
      }
    }
  }

  if (aobj->options & AC3D_OPT_TEXREP) {
    if (aobj->texrep[0]) {
      for (i=0; i<obj->num_surfs; i++) {
	for (j=0; j<obj->surfs[i]->num_elements; j++) {
	  obj->surfs[i]->elements[j].s *= aobj->texrep[0];
	}
      }
      obj->clamp_or_repeat_s = GL_REPEAT;
    } else {
      obj->clamp_or_repeat_s = GL_CLAMP;
    }
    if (aobj->texrep[1]) {
      for (i=0; i<obj->num_surfs; i++) {
	for (j=0; j<obj->surfs[i]->num_elements; j++) {
	  obj->surfs[i]->elements[j].t *= aobj->texrep[1];
	}
      }
      obj->clamp_or_repeat_t = GL_REPEAT;
    } else {
      obj->clamp_or_repeat_t = GL_CLAMP;
    }
  }

  obj->num_kids = aobj->num_kids;
  if (obj->num_kids > 0) {
    obj->kids = gr_new (GrObject *, obj->num_kids);
    if (!obj->kids) {
      obj->num_kids = 0;
      return obj;
    }
    for (i=0; i<obj->num_kids; i++) {
      obj->kids[i] = ac2gr_object (scene, aobj->kids[i], cwd);
      obj->kids[i]->parent = obj;
      obj->kids[i]->scene = scene;
    }
  }

  return obj;
}

static GrScene *
ac2gr_scene (Ac3dScene *ascene, char *cwd)
{
  GrScene *scene;
  int i, j;

  scene = gr_new (GrScene, 1);
  if (!scene)
    return NULL;
  memset (scene, 0, sizeof (*scene));

  scene->num_mats = ascene->num_mats;
  if (scene->num_mats > 0) {
    scene->mats = gr_new (GrMaterial, scene->num_mats);
    if (scene->mats == NULL) {
      scene->num_mats = 0;
      return scene;
    }
    for (i=0; i<scene->num_mats; i++) {
      for (j=0; j<3; j++) {
	scene->mats[i].rgba[j] = ascene->mats[i]->rgb[j];
	scene->mats[i].amb[j] = ascene->mats[i]->amb[j];
	scene->mats[i].dfs[j] = ascene->mats[i]->amb[j];
	scene->mats[i].ems[j] = ascene->mats[i]->ems[j];
	scene->mats[i].spc[j] = ascene->mats[i]->spc[j];
      }
      scene->mats[i].shi = ascene->mats[i]->shi;
      scene->mats[i].rgba[3] = 1.0 - ascene->mats[i]->trans;
      scene->mats[i].amb[3] = 1.0;
      scene->mats[i].ems[3] = 1.0;
      scene->mats[i].dfs[3] = 1.0;
      scene->mats[i].spc[3] = 1.0;
    }
  }

  if (ascene->num_objs != 1) {
    fprintf (stderr, "AC3D file contains more than one top level objects.\n");
    return NULL;
  }

  scene->num_objs = ascene->objs[0]->num_kids;
  if (scene->num_objs > 0) {
    scene->objs = gr_new (GrObject*, scene->num_objs);
    if (scene->objs == NULL) {
      scene->num_objs = 0;
      return scene;
    }
    for (i=0; i<scene->num_objs; i++) {
      scene->objs[i] = ac2gr_object (scene, ascene->objs[0]->kids[i], cwd);
      scene->objs[i]->parent = NULL;
      scene->objs[i]->scene = scene;
    }
  }

  return scene;
}
