/*
 *   GRacer
 *
 *   Copyright (C) 1999 Takashi Matsuda <matsu@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <math.h>
#include "gr_quat.h"

void
gr_quaternion_make_matrix (GrQuaternion *quat, GrMatrix *m)
{
  double q00;
  double q01, q02, q03;
  double q12, q13;
  double q23;

  q00 = quat->q0 * quat->q0;

  q01 = quat->q0 * quat->q.f[0];
  q02 = quat->q0 * quat->q.f[1];
  q03 = quat->q0 * quat->q.f[2];

  q12 = quat->q.f[0] * quat->q.f[1];
  q13 = quat->q.f[0] * quat->q.f[2];

  q23 = quat->q.f[1] * quat->q.f[2];

  m->f[0][0] = 2.0 * (q00 + quat->q.f[0] * quat->q.f[0]) - 1.0;
  m->f[0][1] = 2.0 * (q12 + q03);
  m->f[0][2] = 2.0 * (q13 - q02);
  m->f[0][3] = 0.0;

  m->f[1][0] = 2.0 * (q12 - q01);
  m->f[1][1] = 2.0 * (q00 + quat->q.f[1] * quat->q.f[1]) - 1.0;
  m->f[1][2] = 2.0 * (q23 + q01);
  m->f[1][3] = 0.0;

  m->f[2][0] = 2.0 * (q13 + q02);
  m->f[2][1] = 2.0 * (q23 - q01);
  m->f[2][2] = 2.0 * (q00 + quat->q.f[2] * quat->q.f[2]) - 1.0;
  m->f[2][3] = 0.0;

  m->f[3][0] = 0.0;
  m->f[3][1] = 0.0;
  m->f[3][2] = 0.0;
  m->f[3][3] = 1.0;
}


void
gr_quaternion_normalize (GrQuaternion *d, GrQuaternion *s)
{
  double len = s->q0 * s->q0 +
    	       s->q.f[0] * s->q.f[0] +
    	       s->q.f[1] * s->q.f[1] +
    	       s->q.f[2] * s->q.f[2];

  if (len > 0.0) {
    len = 1.0 / sqrt (len);
    d->q0 = s->q0 * len;
    d->q.f[0] = s->q.f[0] * len;
    d->q.f[1] = s->q.f[1] * len;
    d->q.f[2] = s->q.f[2] * len;
  } else {
    d->q0 = 1.0;
    d->q.f[0] = 0.0;
    d->q.f[1] = 0.0;
    d->q.f[2] = 0.0;
  }
}

void
gr_quaternion_add (GrQuaternion *d, GrQuaternion *a, GrQuaternion *b)
{
  GrQuaternion tmp1, tmp2;
  GrVertex v1, v2, v3;

  gr_quaternion_normalize (&tmp1, a);
  gr_quaternion_normalize (&tmp2, b);

  d->q0 = tmp1.q0 * tmp2.q0 - gr_vertex_product (&tmp1.q, &tmp2.q);
  gr_vertex_scale (&v1, &tmp2.q, tmp1.q0);
  gr_vertex_scale (&v2, &tmp1.q, tmp2.q0);
  gr_vertex_cross (&v3, &tmp1.q, &tmp2.q);
  gr_vertex_add (&d->q, &v1, &v2);
  gr_vertex_add_a (&d->q, &v3);
}

void
gr_quaternion_make_rotation_axis (GrQuaternion *quat, float *angle,
    				  float *x, float *y, float *z)
{
  double len, a;
  GrQuaternion tmp;

  gr_quaternion_normalize (&tmp, quat);

  a = acos (tmp.q0);
  len = sin (a);
  *angle = 2.0 * a;

  if (len > 0) {
    *x = tmp.q.f[0] / len;
    *y = tmp.q.f[1] / len;
    *z = tmp.q.f[2] / len;
  } else {
    *x = 0.0;
    *y = 0.0;
    *z = 1.0;
  }
}
