/*
 *   GRacer
 *
 *   Copyright (C) 1999 Takashi Matsuda <matsu@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sound.h"

#ifdef ENABLE_SOUND
#include <sl.h>

static slScheduler *sched = NULL;

int Sound_Init (Tcl_Interp *interp)
{
  sched = new slScheduler (8000);
  /*
   * check for sound working, otherwise we get a segfault in
   * sched->addSampleEnvelope below if /dev/dsp is not configured
   */
  if (sched->notWorking ()) {
#ifndef __FreeBSD__
    /* XXX this gets a few
     * `gracer in free(): warning: junk pointer, too high to make sense.'
     * don't ask me why. So we'll just leak memory, it should only
     * be allocated once per process...
     */
    delete sched;
#endif
    sched = NULL;
  }
  if (sched) {
    sched->setSafetyMargin (0.2);
  }

  return TCL_OK;
}

int gr_sound_update (void)
{
  if (sched) {
    sched->update ();
  }

  return 1;
}

GrSample gr_sample_new (char *filename)
{
  if (!sched)
    return NULL;

  slSample *s = new slSample (filename, sched);

  return (GrSample *) s;
}

void gr_sample_free (GrSample sample)
{
  slSample *s = (slSample *) sample;

  if (s)
    delete s;
}

void gr_sample_stop (GrSample sample, int magic)
{
  slSample *s = (slSample *) sample;

  if (sched) {
    sched->stopSample (s, magic);
  }
}

void gr_sample_pause (GrSample sample, int magic)
{
  slSample *s = (slSample *) sample;

  if (sched) {
    sched->pauseSample (s, magic);
  }
}

void gr_sample_resume (GrSample sample, int magic)
{
  slSample *s = (slSample *) sample;

  if (sched) {
    sched->resumeSample (s, magic);
  }
}

void gr_sample_adjust_volume (GrSample sample, float volume)
{
  slSample *s = (slSample *) sample;

  if (s) {
    s->adjustVolume (volume);
  }
}

void gr_sample_play (GrSample sample, int magic)
{
  slSample *s = (slSample *) sample;

  if (sched) {
    sched->playSample (s, 1, SL_SAMPLE_ABORT, magic);
  }
}

void gr_sample_loop (GrSample sample, int magic)
{
  slSample *s = (slSample *) sample;

  if (sched) {
    sched->loopSample (s);
  }
}

GrEnvelope gr_envelope_new (int num_step, GrReplayMode mode)
{
  slReplayMode rmode;

  switch (mode) {
  case GR_SAMPLE_LOOP:
    rmode = SL_SAMPLE_LOOP;
    break;

  case GR_SAMPLE_ONE_SHOT:
    rmode = SL_SAMPLE_ONE_SHOT;
    break;

  default:
    return NULL;
  }

  slEnvelope *e = new slEnvelope (num_step, rmode);

  return (GrEnvelope *) e;
}

void gr_envelope_free (GrEnvelope envelope)
{
  slEnvelope *e = (slEnvelope *) envelope;

  if (e)
    delete e;
}

void gr_envelope_set_step (GrEnvelope envelope,
			  int step,
			  float _time,
			  float _value)
{
  slEnvelope *e = (slEnvelope *) envelope;

  if (e) {
    e->setStep (step, _time, _value);
  }
}

void gr_sample_add_envelope	(GrSample sample,
				 int magic,
				 int slot,
				 GrEnvelope envelope,
				 GrEnvelopeType type)
{
  slSample *s = (slSample *) sample;
  slEnvelope *e = (slEnvelope *) envelope;
  slEnvelopeType etype = (slEnvelopeType) type;

  if (!sched)
    return;

  sched->addSampleEnvelope (s, magic, slot, e, etype);
}

#else /* ! ENABLE_SOUND */

int Sound_Init (Tcl_Interp *interp)
{
  return TCL_OK;
}

int gr_sound_update (void)
{
  return 1;
}

GrSample gr_sample_new (char *filename)
{
  return NULL;
}

void gr_sample_free (GrSample sample)
{
}

void gr_sample_stop (GrSample sample, int magic)
{
}

void gr_sample_pause (GrSample sample, int magic)
{
}

void gr_sample_resume (GrSample sample, int magic)
{
}

void gr_sample_adjust_volume (GrSample sample, float volume)
{
}

void gr_sample_play (GrSample sample, int magic)
{
}

void gr_sample_loop (GrSample sample, int magic)
{
}

GrEnvelope gr_envelope_new (int num_step, GrReplayMode mode)
{
  return NULL;
}

void gr_envelope_free (GrEnvelope envelope)
{
}

void gr_envelope_set_step (GrEnvelope envelope,
			  int step,
			  float _time,
			  float _value)
{
}

void gr_sample_add_envelope	(GrSample sample,
				 int magic,
				 int slot,
				 GrEnvelope envelope,
				 GrEnvelopeType type)
{
}

#endif /* ENABLE_SOUND */
